use crate::config::{PAGE_SIZE, PAGE_SIZE_BITS};
use core::fmt::{Debug, Formatter};
use core::slice::from_raw_parts_mut;
use crate::memory::page_table::PageTableEntry;

// 在SV39模式下，物理地址的宽度为56，虚拟地址的宽度为39
pub const PHYSICAL_ADDRESS_WIDTH: usize = 56;
pub const PHYSICAL_PAGE_NUMBER_WIDTH: usize = PHYSICAL_ADDRESS_WIDTH - PAGE_SIZE_BITS;
pub const VIRTUAL_ADDRESS_WIDTH: usize = 39;
pub const VIRTUAL_PAGE_NUMBER_WIDTH: usize = VIRTUAL_ADDRESS_WIDTH - PAGE_SIZE_BITS;

/// 物理地址
#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct PhysicalAddress(pub usize);

impl PhysicalAddress {
    pub fn page_offset(&self) -> usize {
        self.0 & (PAGE_SIZE - 1)
    }

    pub fn floor(&self) -> PhysicalPageNumber {
        PhysicalPageNumber(self.0 / PAGE_SIZE)
    }

    pub fn ceil(&self) -> PhysicalPageNumber {
        if self.0 == 0 {
            PhysicalPageNumber(0)
        } else {
            PhysicalPageNumber((self.0 - 1 + PAGE_SIZE) / PAGE_SIZE)
        }
    }

    pub fn is_aligned(&self) -> bool {
        self.page_offset() == 0
    }
}

impl Debug for PhysicalAddress {
    fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        f.write_fmt(format_args!("Physical address: {:#x}.", self.0))
    }
}

impl From<usize> for PhysicalAddress {
    fn from(value: usize) -> Self {
        Self(value & ((1 << PHYSICAL_ADDRESS_WIDTH) - 1))
    }
}

impl From<PhysicalAddress> for usize {
    fn from(value: PhysicalAddress) -> Self {
        value.0
    }
}

impl From<PhysicalAddress> for PhysicalPageNumber {
    fn from(value: PhysicalAddress) -> Self {
        assert_eq!(value.page_offset(), 0);
        value.floor()
    }
}

impl From<PhysicalPageNumber> for PhysicalAddress {
    fn from(value: PhysicalPageNumber) -> Self {
        Self(value.0 << PAGE_SIZE_BITS)
    }
}

/// 物理页号
#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct PhysicalPageNumber(pub usize);

// 这里的unsafe只是为了绕过编译器的生命周期检查
impl PhysicalPageNumber {
    pub fn get_page_table_entry_array(&self) -> &'static mut [PageTableEntry] {
        let address: PhysicalAddress = (*self).into();
        unsafe {
            // 一个页表的大小是2^12 byte
            // 页表项的大小是8 byte
            // 所有一共有2^9
            from_raw_parts_mut(address.0 as *mut PageTableEntry, 512)
        }
    }

    pub fn get_bytes_array(&self) -> &'static mut [u8] {
        let address: PhysicalAddress = (*self).into();
        unsafe {
            from_raw_parts_mut(address.0 as *mut u8, PAGE_SIZE)
        }
    }

    pub fn get_mut<T>(&self) ->&'static mut T {
        let address: PhysicalAddress = (*self).into();
        unsafe {
            (address.0 as *mut T).as_mut().unwrap()
        }
    }
}

impl Debug for PhysicalPageNumber {
    fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        f.write_fmt(format_args!("Physical page number: {:#x}.", self.0))
    }
}

impl From<usize> for PhysicalPageNumber {
    fn from(value: usize) -> Self {
        Self(value & ((1 << PHYSICAL_PAGE_NUMBER_WIDTH) - 1))
    }
}

impl From<PhysicalPageNumber> for usize {
    fn from(value: PhysicalPageNumber) -> Self {
        value.0
    }
}

/// 虚拟地址
#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct VirtualAddress(pub usize);

impl VirtualAddress {
    pub fn page_offset(&self) -> usize {
        self.0 & (PAGE_SIZE - 1)
    }
    
    pub fn is_aligned(&self) -> bool {
        self.page_offset() == 0
    }
    
    pub fn floor(&self) -> VirtualPageNumber {
        VirtualPageNumber(self.0 / PAGE_SIZE)
    }
    
    pub fn ceil(&self) -> VirtualPageNumber {
        VirtualPageNumber((self.0 - 1 + PAGE_SIZE) / PAGE_SIZE)
    }
}

impl Debug for VirtualAddress {
    fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        f.write_fmt(format_args!("Virtual address: {:#x}.", self.0))
    }
}

impl From<usize> for VirtualAddress {
    fn from(value: usize) -> Self {
        Self(value & ((1 << VIRTUAL_ADDRESS_WIDTH) - 1))
    }
}

impl From<VirtualAddress> for usize {
    fn from(value: VirtualAddress) -> Self {
        value.0
    }
}

/// 虚拟页号
#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct VirtualPageNumber(pub usize);

impl VirtualPageNumber {
    pub fn get_indexes(&self) -> [usize; 3] {
        let mut number = self.0;
        let mut indexes = [0usize; 3];

        for i in 0..3 {
            indexes[i] = number & (1 << 9 - 1);
            number = number >> 9;
        }

        indexes
    }
    
    pub fn step(&mut self) {
        self.0 += 1;
    }
}

impl Debug for VirtualPageNumber {
    fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
        f.write_fmt(format_args!("Virtual page number: {:#x}.", self.0))
    }
}

impl From<usize> for VirtualPageNumber {
    fn from(value: usize) -> Self {
        Self(value & ((1 << VIRTUAL_PAGE_NUMBER_WIDTH) - 1))
    }
}

impl From<VirtualPageNumber> for usize {
    fn from(value: VirtualPageNumber) -> Self {
        value.0
    }
}

#[derive(Copy, Clone)]
pub struct VirtualPageSpan {
    left: VirtualPageNumber,
    right: VirtualPageNumber
}

impl VirtualPageSpan {
    pub fn new(start: VirtualPageNumber, end: VirtualPageNumber) -> Self {
        Self {
            left: start,
            right: end
        }
    }
    
    pub fn get_start(&self) -> VirtualPageNumber {
        self.left
    }
    
    pub fn get_end(&self) -> VirtualPageNumber {
        self.right
    }
}

impl IntoIterator for VirtualPageSpan {
    type Item = VirtualPageNumber;
    type IntoIter = VirtualPageSpanIterator;

    fn into_iter(self) -> Self::IntoIter {
        VirtualPageSpanIterator::new(&self)
    }
}

pub struct VirtualPageSpanIterator {
    current: VirtualPageNumber,
    end: VirtualPageNumber
}

impl VirtualPageSpanIterator {
    fn new(span: &VirtualPageSpan) -> Self {
        Self {
            current: span.left,
            end: span.right
        }
    }
}

impl Iterator for VirtualPageSpanIterator {
    type Item = VirtualPageNumber;
    
    fn next(&mut self) -> Option<Self::Item> {
        if self.current == self.end {
            None
        } else { 
            let t = self.current;
            self.current.step();
            Some(t)
        }
    }
}
