use crate::config::MEMORY_END;
use crate::memory::address::{PhysicalAddress, PhysicalPageNumber};
use crate::sync::single_cell::SingleCell;
use alloc::vec;
use alloc::vec::Vec;
use lazy_static::lazy_static;

trait FrameAllocator {
    fn new() -> Self;

    fn alloc(&mut self) -> Option<PhysicalPageNumber>;

    fn dealloc(&mut self, page_number: PhysicalPageNumber);
}

struct StackedFrameAllocator {
    current: usize,
    end: usize,
    recycled: Vec<usize>,
}

impl FrameAllocator for StackedFrameAllocator {
    fn new() -> Self {
        Self {
            current: 0,
            end: 0,
            recycled: vec![],
        }
    }

    fn alloc(&mut self) -> Option<PhysicalPageNumber> {
        if let Some(page_num) = self.recycled.pop() {
            Some(page_num.into())
        } else {
            if self.current == self.end {
                None
            } else {
                self.current += 1;
                Some((self.current - 1).into())
            }
        }
    }

    fn dealloc(&mut self, page_number: PhysicalPageNumber) {
        let page_number = page_number.0;

        if page_number >= self.current
            || self.recycled.iter().find(|&p| *p == page_number).is_some()
        {
            panic!(
                "Frame physical page number={:#x} has not been allocated.",
                page_number
            );
        }

        self.recycled.push(page_number);
    }
}

impl StackedFrameAllocator {
    pub fn init(&mut self, l: PhysicalPageNumber, r: PhysicalPageNumber) {
        self.current = l.0;
        self.end = r.0;
    }
}

pub struct FrameTracker {
    pub page_number: PhysicalPageNumber
}

impl FrameTracker {
    fn new(page_number: PhysicalPageNumber) -> Self {
        // 清空页帧中的内容
        let byte_array = page_number.get_bytes_array();
        for i in byte_array.as_mut() {
            *i = 0;
        }
        
        Self {
            page_number
        }
    }
}

impl Drop for FrameTracker {
    fn drop(&mut self) {
        FRAME_ALLOCATOR.exclusive_borrow().dealloc(self.page_number);
    }
}

lazy_static! {
    static ref FRAME_ALLOCATOR: SingleCell<StackedFrameAllocator> =
        SingleCell::new(StackedFrameAllocator::new());
}

pub fn initialize_frame_allocator() {
    extern "C" {
        fn ekernel();
    }

    FRAME_ALLOCATOR.exclusive_borrow().init(
        PhysicalAddress::from(ekernel as usize).ceil(),
        PhysicalAddress::from(MEMORY_END).floor(),
    )
}       

pub fn allocate_frame() -> Option<FrameTracker> {
    FRAME_ALLOCATOR.exclusive_borrow().alloc().map(|p| FrameTracker::new(p))
}

