use alloc::collections::BTreeMap;
use alloc::vec;
use alloc::vec::Vec;
use bitflags::bitflags;
use crate::config::PAGE_SIZE;
use crate::memory::address::{PhysicalPageNumber, VirtualAddress, VirtualPageNumber, VirtualPageSpan};
use crate::memory::frame_allocator::{allocate_frame, FrameTracker};
use crate::memory::page_table::{PageTable, PageTableEntryFlags};

/// 虚拟地址的映射方式
#[derive(Copy, Clone, Eq, PartialEq, Debug)]
pub enum MapType {
    /// 恒等映射
    Identical,
    /// 新分配物理页帧
    Framed
}

bitflags! {
    /// 逻辑段的权限控制
    /// 只是PageTableEntryFlags的子集
    /// 其他的标志位对于权限控制无用，只和地址转换有关
    pub struct MapPermission: u8 {
        const R = 1 << 1;
        const W = 1 << 2;
        const X = 1 << 3;
        const U = 1 << 4;
    }
}

/// 一段地址连续的虚拟内存
pub struct MapArea {
    virtual_page_span: VirtualPageSpan,
    data_frames: BTreeMap<VirtualPageNumber, FrameTracker>,
    map_type: MapType,
    map_permission: MapPermission
}

impl MapArea {
    pub fn new(
        start: VirtualAddress,
        end: VirtualAddress,
        map_type: MapType,
        map_permission: MapPermission
    ) -> Self {
        let start = start.floor();
        let end = end.ceil();
        
        Self {
            virtual_page_span: VirtualPageSpan::new(start, end),
            data_frames: BTreeMap::new(),
            map_type,
            map_permission
        }
    }
    
    pub fn map(&mut self, page_table: &mut PageTable) {
        for page_number in self.virtual_page_span {
            self.map_one(page_table, page_number);
        }
    }
    
    pub fn unmap(&mut self, page_table: &mut PageTable) {
        for page_number in self.virtual_page_span {
            self.unmap_one(page_table, page_number);
        }
    }
    
    /// Copy data into this MapArea
    /// The data is start-aligned and shorter than MapArea!
    pub fn copy_data(&mut self, page_table: &mut PageTable, data: &[u8]) {
        assert_eq!(self.map_type, MapType::Framed);
        let mut start = 0;
        let mut current_page_number = self.virtual_page_span.get_start();
        let length = data.len();

        loop {
            let source = &data[start..length.min(start + PAGE_SIZE)];
            let destination = &mut page_table
                .translate(current_page_number)
                .unwrap()
                .physical_page_numer()
                .get_bytes_array()[..source.len()];
            
            destination.copy_from_slice(source);
            start += PAGE_SIZE;
            if start >= length {
                break
            }
            
            // 自增
            current_page_number.step()
        }
    }
    
    fn map_one(&mut self, page_table: &mut PageTable, virtual_page_number: VirtualPageNumber) {
        let physical_page_number = match self.map_type {
            MapType::Identical => {
                // 使用和虚拟页号一样的物理页号
                PhysicalPageNumber(virtual_page_number.0)
            },
            MapType::Framed => {
                // 分配新的物理页帧
                let frame = allocate_frame().unwrap();
                let frame_number = frame.page_number;
                self.data_frames.insert(virtual_page_number, frame);
                frame_number
            }
        };
        let flags = PageTableEntryFlags::from_bits(self.map_permission.bits()).unwrap();
        page_table.map(virtual_page_number, physical_page_number, flags);
    }
    
    fn unmap_one(&mut self, page_table: &mut PageTable, virtual_page_number: VirtualPageNumber) {
        match self.map_type {
            MapType::Framed => {
                self.data_frames.remove(&virtual_page_number);
            },
            _ => {}
        }
        page_table.unmap(virtual_page_number);
    }
}

pub struct MemorySet {
    page_table: PageTable,
    areas: Vec<MapArea>
}

impl MemorySet {
    pub fn new_bare() -> Self {
        Self {
            page_table: PageTable::new(),
            areas: vec![]
        }
    }
    
    fn push(&mut self, mut map_area: MapArea, data: Option<&[u8]>) {
        
    }
}




